
package jartege ; 

import java.lang.reflect.* ; 

/**
 * Classe des appels.
 */

class Appel extends Expression {
   
   // L'operation de cet appel
   private Operation operation ; 

   // Le parametre implicite de l'appel
   private Expression paramImpl ; 

   // Le tableau des parametres explicites de l'appel
   private Expression[] tabParam ; 

   /**
    * Constructeur.
    */
   Appel(Operation operation, Expression paramImpl, Expression[] tabParam) {
      super() ; 
      this.operation = operation ; 
      this.paramImpl = paramImpl ; 
      this.tabParam = tabParam ; 
   }

   /**
    * L'operation associe  cet appel.
    */
   Operation getOperation() {
      return operation ; 
   }

   /**
    * Le paramtre implicite de cet appel.
    */
   Expression getParamImpl() {
      return paramImpl ; 
   }

   /**
    * L'objet java denote par le parametre implicite de cet appel.
    */
   Object getValeurParamImpl() {
      if (paramImpl != null) {
         return paramImpl.getValeur() ; 
      } else {
         return null ; 
      }
   }

   /**
    * Le tableau des paramtres de cet appel.
    */
   Expression[] getTabParam() {
      return tabParam ; 
   }

   /**
    * Le tableau d'objets Java denote par les parametres de cet appel.
    */
   Object[] getValeurTabParam() {
      Object[] result = new Object[tabParam.length] ; 
      for (int i = 0 ; i < tabParam.length ; i++) {
         result[i] = tabParam[i].getValeur() ; 
      }
      return result ; 
   }

   /**
    * Affiche un message d'erreur.
    */
   void messageErreur() {
      System.err.println("Erreur dans la generation d'un appel de " +
            operation.nom + ".") ;
   }

   /**
    * Evalue cet appel.
    */
   public Object eval() {
      Trace.afficher(4, "Appel.eval() ->") ; 
      Trace.afficher(4, toString()) ; 

      // Le resultat
      Object res = null ; 
      
      try {
         res = operation.eval(this) ; 
      } catch (InstantiationException except) {
         messageErreur() ; 
         System.err.println(
            "Tentative d'instanciation d'une classe abstraite") ; 
         System.err.println(
            "Les fichiers .class ne sont peut-etre pas a jour.") ; 
         except.printStackTrace() ; 
      } catch (IllegalAccessException except) {
         System.err.println("Erreur dans la generation d'un appel de " + 
            operation.nom + ".") ; 
         System.err.println("Operation inaccessible.") ;
         System.err.println("  - La classe ou l'interface " +
            operation.laClasse.getNomComplet() + 
            " n'est pas accessible (devrait etre publique)") ; 
         System.err.println("  - L'operation " + 
            operation.getNomComplet() +
            " n'est pas accessible (devrait etre publique)") ;
      } catch (IllegalArgumentException except) {
         System.err.println("Erreur dans la generation d'un appel de " + 
            operation.nom + ".") ; 
         System.err.println(
            "Les fichiers .class ne sont peut-etre pas a jour.") ; 
         System.err.println("Parametres incorrects : tabParam.length = " +
            tabParam.length) ;
         for (int i = 0 ; i < tabParam.length ; i++) {
            System.out.println("   tabParam[" + i + "] = " + tabParam[i]) ;
         }
         System.err.println("\nAppel : " + operation.toString(this)) ;
         throw except ;
      } catch (InvocationTargetException except) {
         res = new ResultatException(except.getTargetException()) ;
      }
      return res ; 
   }

   /**
    * Un appel sous forme d'une chaine de caracteres.
    */

   public String toString() {
      return operation.toString(this) ; 
   }

}

