package jartege ; 

import java.util.* ; 
import java.io.* ; 
import jartege.* ; 
import java.lang.reflect.* ; 

/**
 * Class that allows one to generate random tests for a specified set of 
 * classes. <br>
 * A typical use of this class is as follows: 
 * <code>
 * <ul>
 *    public static void main(String[] args) { <br>
 *    <ul>
 *       ClassTester t = new ClassTester(args) ; <br>
 *       t.addClass("packageName.ClassA") ; <br>
 *       t.addClass("packageName.ClassB") ; <br>
 *       t.changeClassWeight("packageName.ClassA", 9) ; <br>
 *       t.changeCreationProbability("packageName.ClassA", 
 *       <ul>
 *          new ThresholdProbability(1)) ; 
 *       </ul>
 *       t.addImportedPackage("packageName") ; <br>
 *       t.generate("TestForAandB", 10, 50) ; <br>
 *    </ul>
 *    }
 * </ul>
 * </code>
 *
 * A <em>weight</em> is positive or null double value associated with 
 * each class and each visible operation of 
 * class (constructors, methods and operations get and set for each visible 
 * field).
 * By default, all weights are equal to 1.
 * The program generates a method call according to the specified weights.
 * Weights can be changed with change weight methods.
 * The following statement change the weight of 
 * <code>packageName.ClassA</code> to 9.
 *
 * <code>
 * <ul>
 *    t.changeClassWeight("packageName.ClassA", 9) ; 
 * </ul>
 * </code>
 *
 * A <em>creation probability function</em> is associated with each class under
 * test. This function commands the creation of new objects of the 
 * class, according to the number of already created objects. <br>
 * <br>
 * The statement
 * <code>
 * <ul>
 *    t.changeCreationProbability("packageName.ClassA", <br>
 *    <ul>
 *       new ThresholdProbability(1)) ; 
 *    </ul>
 * </ul>
 * </code>
 * changes the creation threshold function of class <code>ClassA</code>
 * to the specified probability function, 
 * which allows the creation of at most one object of <code>ClassA</code>. <br>
 * <br>
 *
 * The statement
 * <code>
 * <ul>
 *       t.addImportedPackage("packageName") ;
 * </ul>
 * </code>
 * add the package <code>packageName</code> to the set of imported packages.
 * For each imported package, the generator generates a line
 * <code>
 * <ul>
 *    import <em>packageName</em>.* ; 
 * </ul>
 * </code>
 * in the generated test file.<br>
 * <br>
 *
 * The statement
 * <code>
 * <ul>
 *    t.generate("TestForAandB", 10, 50) ; <br>
 * </ul>
 * </code>
 * generates a test file <code>"TestForAandB.java"</code>.
 * This file consists of a class called <code>"TestForAandB"</code>
 * which contains 10 test methods. For each test method, the tool
 * tries to generate 50 method calls in classes <code>ClassA</code>
 * and <code>ClassB</code>. For each generated call, <code>ClassA</code>
 * will be choosen with a probablity of 0.9 while <code>ClassB</code>
 * will be chosen with a probability of 0.1. <br>
 *
 * @author Catherine Oriat
 */

public class ClassTester {

   // La liste des classes que l'on cherche  tester
   List lesClasses ; 

   // The set of classes under test
   private Set theClasses ; 

   //
   Dictionnaire leDictionnaire ; 

   // Le nom du paquetage dans lequel on gnre le fichier
   private String packageName ; 

   // Le nom du rpertoire dans lequel on gnre le fichier de test
   private String dirName ; 

   // L'ensemble des noms des paquetages imports
   private Set importedPackages ; 

   // L'ensemble des classes importes
   private Set importedClasses ; 

   // The classe to extend
   private Class toExtendClass ; 

   private Constructor consToExtend ; 

   /**
    * Constructs a new class tester. 
    *
    * @param args the arguments of the command line.
    */
    
   public ClassTester(String[] args) {
      System.out.println("\nJartege v0.04") ; 
      Argument.decode(args) ; 
      lesClasses = new Vector() ; 
      theClasses = new HashSet() ; 
      leDictionnaire = new Dictionnaire(this) ; 
      dirName = "" ; 
      importedPackages = new HashSet() ; 
      importedClasses = new HashSet() ; 
      toExtendClass = null ; 
      consToExtend = null ; 
   }

   /**
    * Constructs a new class tester. <br>
    */

   public ClassTester() {
      this(new String[0]) ; 
   }

   /**
    * Add the specified class to the context.
    */

   public void addContext(String className) {
      Classe classe = leDictionnaire.getClasse(className) ; 
   }


   /**
    * Returns the package name of the generated test file.
    * Returns null if the package name has not been previously set.
    */

   public String getPackage() {
      return packageName ; 
   }

   /**
    * Reinitialises the context.
    */

   public void resetContext() {
      leDictionnaire = new Dictionnaire(this) ; 
   }


   /**
    * Sets the package name of the generated test file to the specified 
    * package name.
    *
    * @param packageName the name of the package in which the 
    *                    test file is generated.
    */

   public void setPackage(String packageName) {
      if (packageName == null || packageName.equals("")) {
         ErrorMessage.incorrectPackageName(packageName) ; 
      }
      this.packageName = packageName ; 
   }


   /**
    * Sets the directory in which the test file is generated to the
    * specified directory name.
    * If the directory is not set, then the file is generated in the
    * current directory.
    *
    * @param dirName the name of the directory in which the test file
    *        is generated.
    */

   public void setDir(String dirName) {
      this.dirName = dirName ; 
   }

   /**
    * Returns the directory in which the test file is generated.
    * Returns null if the directory has not been previously set.
    * @see #setDir
    */

   public String getDir() {
      return dirName ; 
   }

   /**
    * Sets the ForJUnit flag to the specified boolean value.
    * If the ForJUnit flag is true, then the generator will 
    * generate a test for JUnit, otherwise the generator 
    * generate a stand alone class test.
    * By default, the ForJUnit flag is false.
    */

   public void setForJUnit(boolean b) {
      Constantes.FOR_JUNIT = b ; 
   }

   /**
    * Returns the ForJUnit flag value.
    */

   public boolean getForJUnit() {
      return Constantes.FOR_JUNIT ; 
   }

   /**
    * Adds the specified package to the set of imported packages.
    * The test generator will the generate a line <br>
    * <code>
    *    import <em>packageName</em>.* ; <br>
    * </code>
    * in the generated test file.
    *
    * @param packageName the name of the package that has to be imported. 
    * @see #addImportedClass
    */

   public void addImportedPackage(String packageName) {
      importedPackages.add(packageName) ; 
   }

   /**
    * Returns the set of imported packages.
    * Each element of the set is of type String.
    */

   public Set getImportedPackages() {
      return importedPackages ; 
   } 

   /**
    * Adds the specified class to the set of imported classes.
    * The test generator will generate a line  <br>
    * <code>
    *    import <em>className</em> ; <br>
    * </code>
    * in the generated test file.
    *
    * @param className the name of the class that has to be imported.
    * @see #addImportedPackage
    */

   public void addImportedClass(String className) {
      importedClasses.add(Util.forName(className)) ; 
   }

   /**
    * Returns the set of imported classes.
    * Each element of the set is of type Class.
    */

   public Set getImportedClasses() {
      return importedClasses ; 
   }

   /**
    * Defines the class that the generator extends. 
    * If the setToExtendClass method is not called, then the generator
    * produces a class which extends java.lang.Object.
    * When the ForJUnit flag is true, the class that the generator extends 
    * must extend the JUnit class TestCase.
    * The class that the generator extends can be unset with the
    * <code>unsetToExtendClass()</code> method.
    *
    * @param toExtendClassName the name of the class that will be extended
    *        by the generator.
    * @see #unsetToExtendClass
    */


   public void setToExtendClass(String toExtendClassName) {
      toExtendClass = Util.forName(toExtendClassName) ; 
      if (Modifier.isAbstract(toExtendClass.getModifiers())) {
         ErrorMessage.shouldNotBeAbstract(toExtendClassName) ; 
      }
      consToExtend = null ; 
      Set constructors = Constituant.getAllVisibleConstructors
                            (toExtendClass, toExtendClass) ;
      Iterator it = constructors.iterator() ;
      while (it.hasNext() && consToExtend == null) {
         Constructor cons = (Constructor) it.next() ;
         if (cons.getParameterTypes().length == 0) {
            consToExtend = cons ;
         }
      }
      if (consToExtend == null) {
         ErrorMessage.noConstructorInClass(toExtendClass.getName()) ; 
      }
      consToExtend.setAccessible(true) ;
   }

   Constructor getConsToExtend() {
      return consToExtend ; 
   }

   /**
    * Returns the class that the generated test class extends.
    */

   public Class getToExtendClass() {
      return toExtendClass ; 
   }

   /**
    * Unsets the class that the generated test class extends. 
    *
    * @see #setToExtendClass
    */

   public void unsetToExtendClass() {
      toExtendClass = null ; 
   }


   /**
    * Normal creation probability function.
    * The probability function is defined by 
    * <pre>
    *    theFunction(n) = 
    *       1.0 / (1.0 + n) ; 
    * </pre>
    */


   public static CreationProbability creationProbabilityNormal = 
      new CreationProbability() {
         public double theFunction(int nb) {
            return 1.0 / (1.0 + nb) ; 
         }
      } ; 

   /**
    * Probability function that allows the creation of few objects.
    * The probability function is defined by
    * <pre>
    *    theFunction(n) =
    *       1.0 / ((1.0 + n) * (1.0 + n)) ;
    * </pre>
    */

   public static CreationProbability creationProbabilityFew = 
      new CreationProbability() {
         public double theFunction(int nb) {
            return 1.0 / ((1.0 + nb) * (1.0 + nb)) ; 
         }
      } ; 

   /**
    * Probability function that allows the creation of many objects.
    * The probability function is defined by
    * <pre>
    *    theFunction(n) =
    *       1.0 / Math.sqrt(1.0 + n) ;
    * </pre>
    */

   public static CreationProbability creationProbabilityMany = 
      new CreationProbability() {
         public double theFunction(int nb) {
            return 1.0 / Math.sqrt(1.0 + nb) ; 
         }
      } ; 

   
   /**
    * Adds the class with the specified name to the set of classes 
    * under test.
    *
    * @param className the name of the class which is added in the set 
    *        of classes under test.
    */

   public void addClass(String className) {
      Class clas = Util.forName(className) ; 
      if (!theClasses.contains(clas)) {
         theClasses.add(clas) ; 
         Classe classe = leDictionnaire.getClasse(clas) ; 
         lesClasses.add(classe) ; 
      }
   }

   /**
    * Remove the class with the specified name to the set of classes
    * under test.
    *
    * @param className the name of the class which is removed of the set
    *        of classes under test.
    */

   public void removeClass(String className) {
      Class clas = Util.forName(className) ;
      if (theClasses.contains(clas)) {
         theClasses.remove(clas) ; 
         Classe classe = leDictionnaire.getClasse(clas) ; 
         lesClasses.remove(classe) ; 
      }
   }

   /**
    * Removes all classes under test in this class tester.
    */

   public void removeAllClasses() {
      theClasses = new HashSet() ; 
      lesClasses = new Vector() ; 
   }


   /**
    * Returns the set of classes under test of this class tester.
    */

   public Set getClasses() {
      return theClasses ; 
   }

   List lesClasses() {
      return lesClasses ; 
   }

   /**
    * Changes the weight of class with the specified name to the 
    * specified weight.
    */

   public void changeClassWeight(String className, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      Classe classe = leDictionnaire.getClasse(className) ; 
      classe.setPoids(weight) ; 
   }

   /**
    * In class with the specified name, changes all methods weight to 
    * the specified weight.
    */

   public void changeAllMethodsWeight(String className, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      Classe classe = leDictionnaire.getClasse(className) ; 
      List op = classe.lesOperations ; 
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ;
         o.setPoids(weight) ; 
      }
   }

   /**
    * In class with the specified name, changes all constructors weight to
    * the specified weight.
    */

   public void changeAllConstructorsWeight(String className, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      Classe classe = leDictionnaire.getClasse(className) ;
      List op = classe.lesConstructeurs ;
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ;
         o.setPoids(weight) ;
      }
   }

   /**
    * In class with the specified name <code>className</code>, changes the 
    * weight of all methods with the specified name 
    * <code>methodName</code> to the specified weight.
    */

   public void changeMethodWeight
         (String className, String methodName, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      boolean ok = false ; 
      String meth = Chaine.nom(methodName) ; 
      Classe classe = leDictionnaire.getClasse(className) ; 
      List op = classe.lesOperations ; 
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ; 
         if (o.getNom().equals(meth)) {
            o.setPoids(weight) ; 
            ok = true ; 
         }
      }
      if (!ok) {
         ErrorMessage.methodNotFound(
            "changeMethodWeight", methodName, className) ; 
      }
   }

   /**
    * In class with the specified name <code>className</code>, changes the
    * weight of the method with the specified name <code>methodName</code>
    * and specified signature to the specified weight.
    */

   public void changeMethodWeight
         (String className, String methodName, 
          String[] theSignature, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      boolean ok = false ; 
      Classe classe = leDictionnaire.getClasse(className) ; 
      String meth = Chaine.nom(methodName) ; 
      Class[] theSig = new Class[theSignature.length] ; 
      for (int i = 0 ; i < theSignature.length ; i++) {
         theSig[i] = leDictionnaire.getClasse(theSignature[i]).theClass ; 
      }
      List op = classe.lesOperations ; 
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ; 
         if (o.getNom().equals(meth) && 
             Arrays.equals(o.typeParametres(), theSig)) {
            o.setPoids(weight) ; 
            ok = true ; 
         }
      } 
      if (!ok) {
         ErrorMessage.methodNotFound(
            "changeMethodWeight", methodName, theSignature, className) ; 
      }
   }

   /**
    * In class with the specified name <code>className</code>, changes the
    * weight of the constructor with the specified name 
    * <code>constructorName</code>
    * and specified signature to the specified weight.
    */

   public void changeConstructorWeight
         (String constructorName, String[] theSignature, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      boolean ok = false ; 
      Classe classe = leDictionnaire.getClasse(constructorName) ; 
      String constr = Chaine.nom(constructorName) ; 
      Class[] theSig = new Class[theSignature.length] ; 
      for (int i = 0 ; i < theSignature.length ; i++) {
         theSig[i] = leDictionnaire.getClasse(theSignature[i]).theClass ; 
      }
      List op = classe.lesConstructeurs ; 
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ; 
         if (o.getNom().equals(constr) && 
             Arrays.equals(o.typeParametres(), theSig)) {
            o.setPoids(weight) ; 
            ok = true ; 
         }
      } 
      if (!ok) {
         ErrorMessage.constructorNotFound(
            "changeConstructorWeight", constructorName, theSignature) ; 
      }
   }

   private boolean changeGetSetWeight
          (String className, String fieldName, double weight, boolean isGet) {
      boolean ok = false ;
      Classe classe = leDictionnaire.getClasse(className) ;
      String champ = Chaine.nom(fieldName) ;
      List op = classe.lesOperations ;
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ;
         if (isGet) {
            if (o instanceof OperationGet) {
               OperationGet o2 = (OperationGet) o ;
               if (o2.theField.getName().equals(champ)) {
                  o.setPoids(weight) ;
                  ok = true ;
               }
            }
         } else {
            if (o instanceof OperationSet) {
               OperationSet o2 = (OperationSet) o ;
               if (o2.theField.getName().equals(champ)) {
                  o.setPoids(weight) ;
                  ok = true ;
               }
            }
         }
      }
      return ok ; 
   }

   /**
    * Changes the weight of the get operation associated with the 
    * specified field in the specified class to the specified weight.
    */

   public void changeGetWeight
          (String className, String fieldName, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      boolean ok ; 
      ok = changeGetSetWeight(className, fieldName, weight, true) ; 
      if (!ok) {
         ErrorMessage.fieldNotFound("changeGetWeight", fieldName, className) ;
      }
   }

   /**
    * Changes the weight of the set operation associated with the 
    * specified field in the specified class to the specified weight.
    */

   public void changeSetWeight
          (String className, String fieldName, double weight) {
      if (weight < 0) {
         ErrorMessage.negativeWeight("changeClassWeight", weight) ; 
      }
      boolean ok ; 
      ok = changeGetSetWeight(className, fieldName, weight, false) ;
      if (!ok) {
         ErrorMessage.fieldNotFound("changeSetWeight", fieldName, className) ; 
      }
   }

   private double getGetSetWeight
          (String className, String fieldName, boolean isGet) {
      Classe classe = leDictionnaire.getClasse(className) ;
      String champ = Chaine.nom(fieldName) ;
      List op = classe.lesOperations ;
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ;
         if (isGet) {
            if (o instanceof OperationGet) {
               OperationGet o2 = (OperationGet) o ;
               if (o2.theField.getName().equals(champ)) {
                  return o.getPoids() ; 
               }
            }
         } else {
            if (o instanceof OperationSet) {
               OperationSet o2 = (OperationSet) o ;
               if (o2.theField.getName().equals(champ)) {
                  return o.getPoids() ;
               }
            }
         }
      }
      // Operation not found
      if (isGet) {
         ErrorMessage.getNotFound("getGetWeight", fieldName, className) ; 
      } else {
         ErrorMessage.setNotFound("getSetWeight", fieldName, className) ;
      }
      return 0 ;
   }
      

   /**
    * Gets the weight of the get operation associated with the specified
    * field in the specified class.
    * 
    * @param className name of the class.
    * @param fieldName field name in the specified class.
    */

   public double getGetWeight
          (String className, String fieldName) {
      return getGetSetWeight(className, fieldName, true) ; 
   }

   /**
    * Gets the weight of the get operation associated with the specified
    * field in the specified class.
    * 
    * @param className name of the class.
    * @param fieldName field name in the specified class.
    */

   public double getSetWeight
          (String className, String fieldName) {
      return getGetSetWeight(className, fieldName, false) ; 
   }

   /**
    * Changes the creation probability function of the specified class 
    * to the specified function.
    */

   public void changeCreationProbability
         (String className, CreationProbability function) {
      Classe classe = leDictionnaire.getClasse(className) ; 
      classe.objectGen.setCreationProbability(function) ; 
   }

   /**
    * Returns the creation probability associated with the specified class.
    */

   public CreationProbability getCreationProbability
         (String className) {
      Classe classe = leDictionnaire.getClasse(className) ;
      return classe.objectGen.getCreationProbability() ;  
   }

   /**
    * Sets the storeResult flag of all methods in the specified class
    * to the specified boolean value.
    */

   public void setAllMethodsStoreResult 
      (String className, boolean storeResult) {
      Classe classe = leDictionnaire.getClasse(className) ;
      List op = classe.lesOperations ;
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ;
         o.setStoreResult(storeResult) ; 
      }
   }

   /**
    * Sets the storeResult flag of method with the specified name in
    * the specified class to the specified boolean value.
    */

   public void setMethodStoreResult
      (String className, String methodName, boolean storeResult) {

      boolean ok = false ; 
      String meth = Chaine.nom(methodName) ; 
      Classe classe = leDictionnaire.getClasse(className) ; 
      List op = classe.lesOperations ; 
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ; 
         if (o.getNom().equals(meth)) {
            o.setStoreResult(storeResult) ; 
            ok = true ; 
         }
      }
      if (!ok) {
         ErrorMessage.methodNotFound(
            "setMethodStoreResult", methodName, className) ; 
      }
   }


   /**
    * Indicates that results of method calls with the specified name 
    * with the specified signature in
    * the specified class should not be stored in variables.
    */

   public void setMethodStoreResult
         (String className, 
          String methodName, 
          String[] signature, 
          boolean storeResult) {

      boolean ok = false ; 
      Classe classe = leDictionnaire.getClasse(className) ; 
      String meth = Chaine.nom(methodName) ; 
      Class[] theSig = new Class[signature.length] ; 
      for (int i = 0 ; i < signature.length ; i++) {
         theSig[i] = leDictionnaire.getClasse(signature[i]).theClass ; 
      }
      List op = classe.lesOperations ; 
      for (int i = 0 ; i < op.size() ; i++) {
         Operation o = (Operation) (op.get(i)) ; 
         if (o.getNom().equals(meth) && 
             Arrays.equals(o.typeParametres(), theSig)) {
            o.setStoreResult(storeResult) ; 
            ok = true ; 
         }
      } 
      if (!ok) {
         ErrorMessage.methodNotFound(
            "setMethodStoreResult", methodName, signature, className) ; 
      }
   }


   /**
    * Change the object generator associated with the class with 
    * the specified name <code>className</code> to the specified 
    * object generator.
    *
    * @param className the name of the class.
    * @param objectGen the object generator.
    */

   public void changeObjectGenerator
         (String className, ObjectGenerator objectGen) {
      Classe classe = leDictionnaire.getClasse(className) ;
      classe.objectGen = objectGen ;
      objectGen.setClasse(classe) ; 
   }




   /**
    * Generates a test class with the specified name 
    * <code>className</code>.
    *
    * @param className the name of the generated test class.
    * @param numberOfTests the number of test methods generated in the 
    *        class.
    * @param numberOfMethodCalls the number of method calls the generator
    *        tries to generate for each test method.
    */

   public void generate
      (String className,
       int numberOfTests, 
       int numberOfMethodCalls) {

      boolean erreur = false ; 
      Iterator it = theClasses.iterator() ; 
      while (it.hasNext()) {
         Class clas = (Class) it.next() ; 
         if (!Modifier.isPublic(clas.getModifiers())) {
            Package thePackage = clas.getPackage() ; 
            if (thePackage == null) {
               if (packageName != null) {
                  ErrorMessage.classShouldBePublic(clas.getName()) ;
                  erreur = true ; 
               }
            } else if (packageName == null || 
                       !packageName.equals(thePackage.getName())) {
               ErrorMessage.classShouldBePublic(clas.getName()) ;
               erreur = true ; 
            }
         }
      }
      if (erreur) {
         System.exit(1) ; 
      }


      
      Generateur gen = new Generateur(this, className) ; 

      gen.generate
         (numberOfTests, numberOfMethodCalls) ; 
   }

   /**
    * Generates a test class with the specified name 
    * <code>className</code>.
    * This is equivalent to generate(className, 1, numberOfMethodCalls)
    *
    * @param className the name of the generated test class.
    * @param numberOfMethodCalls the number of method calls the generator
    *        tries to generate.
    */

   public void generate (String className, int numberOfMethodCalls) {
      generate(className, 1, numberOfMethodCalls) ; 
   }

}

