package jartege ; 

/**
 * The class ConstantProbability allows to create creation
 * probability functions which are constant. 
 * A creation probablity function defines
 * the probability of creation of an object in function of the number
 * of already created objects. This probability is constant.<br>
 * <code> 
 * <ul>
 *    theFunction(nbCreatedObjects) = <em>constant value</em>.
 * </ul>
 * </code>
 *
 * @author Catherine Oriat
 */

public class ConstantProbability implements CreationProbability {

   private final double constant ; 

   /**
    * Constructs a new creation probability function defined by
    * the specified constant.
    *
    * @param constant the constant value of the constant probability function.
    */

   public ConstantProbability(double constant) {
      this.constant = constant ; 
   }

   /**
    * The creation probability function. 
    * <code>
    * <ul>
    *    theFunction(nbCreatedObjects) = <em>constant value</em>.
    * </ul>
    * </code>
    */

   public double theFunction(int nbCreatedObjects) {
      return constant ; 
   }

   /**
    * Returns the constant probability of this constant creation 
    * probability function.
    */

   public double getConstant() {
      return constant ; 
   }
}

